/*  File: analyser-gap.cc
 *  Project: Deductor
 *  Created by Zoltán Nagy on 9/8/13.
 *  Modified by D. Soper on 23 February 2019.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   local headers  */
#include "analyzer-gap.h"
#include <math.h>

/* Approximate fits to d sigma/dpT dDelta y for jets with R = 0.4
   and Y < Delta y < Y+1.
 */
static double cross_section12(double pT)
{
  double x  = log(pT/40.0);
  double x2 = x*x;
  double x3 = x2*x;
  double x4 = x3*x;
  double y = 7.20941614072688 - 4.870105310740897*x
             - 0.08370266068449551*x2 - 0.02673282428530197*x3
             - 0.005862638784321819*x4;
  return exp(y);
}

static double cross_section23(double pT)
{
  double x  = log(pT/40.0);
  double x2 = x*x;
  double x3 = x2*x;
  double x4 = x3*x;
  double y = 7.010703838639867 - 5.037673805021665*x
             - 0.07608981873370878*x2 - 0.039642089691458986*x3
             - 0.006840026905954689*x4;
  return exp(y);
}

static double cross_section34(double pT)
{
  double x  = log(pT/40.0);
  double x2 = x*x;
  double x3 = x2*x;
  double x4 = x3*x;
  double y = 6.540428361856565 - 5.181632649374608*x
             - 0.23447128297868686*x2 + 0.04360267744938224*x3
            - 0.03026290787820582*x4;
  return exp(y);
}

static double cross_section45(double pT)
{
  double x  = log(pT/40.0);
  double x2 = x*x;
  double x3 = x2*x;
  double x4 = x3*x;
  double y = 5.8885654403589385 - 5.4099190821997105*x
             - 0.38198720808352815*x2 + 0.1518077960975368*x3
             - 0.07566649762912006*x4;
  return exp(y);
}

static double cross_section56(double pT)
{
  double x  = log(pT/40.0);
  double x2 = x*x;
  double x3 = x2*x;
  double x4 = x3*x;
  double y = 5.069725050386662 - 5.311590539168829*x
             - 1.3837909801758739*x2 + 0.8709271542376557*x3
             - 0.2863460817587929*x4;
  return exp(y);
}

void UserGap::initfunc()
{
  /*  Create histograms to calculate the fraction of two jet events with a gap
   *  as a function of the PTcut for the gap.
   *  We demand that there are two jets with PT > PTjet GeV
   *  one with Y < - Ygap and one with Y > + Ygap.
   *  Then there can be no jet in the gap -Ygap < Y < +Ygap
   *  with PT > PTcut
   */
  auto bins = spacing<hist1d>::manual({300,400}); // Just one bin.

  /* HistArr was defined as a 6 dimensional array of histograms  */
  HistArr::phys(1, "Gap cross sections, 1 < Delta y < 2", bins);
  HistArr::phys(2, "Gap cross sections, 2 < Delta y < 3", bins);
  HistArr::phys(3, "Gap cross sections, 3 < Delta y < 4", bins);
  HistArr::phys(4, "Gap cross sections, 4 < Delta y < 5", bins);
  HistArr::phys(5, "Gap cross sections, 5 < Delta y < 6", bins);
  
  cout << "UserGap is using " << jetDef.description() << "."<< endl;
  cout << "We have set sqrt(s) = 13 TeV." << jetDef.description() << endl;
  cout << "Gap is between jets, with PTcut = 20 GeV." << endl;
}


void UserGap::userfunc(const shower_stage<hhc>&,          // hard
                       const shower_stage<hhc>&,          // afterthreshold
                       const shower_stage<hhc>& showered)
{
  constexpr double tonb = 389379.338;
  constexpr double tiny = 1.0e-6;
  
  /* showered event */
  auto& p = showered.state;
  /* Weights */
  auto wi = weight_info(showered);
  /*   This is a resonable breakdown of the full weight
   *     0 => Standard weight (no delta, no threshold)
   *     1 => Standard * (Standard Threshold)
   *     2 => (Standard + Delta) (no threshold)
   *     3 => (Standard + Delta + DeltaThreshold)*(Standard Threshold)
   *     4 => (Standard + Delta + Phase) (no threshold)
   *     5 => (Standard + Delta + Phase + DeltaThreshold)*(Standard Threshold)
   */
   std::array<double, 6u> w{{tonb*wi[1], tonb*wi[1],
                             tonb*wi[3], tonb*wi[3],
                             tonb*wi[5], tonb*wi[5]}};
  /* These will become
   {LC+ only with gap, LC+ only total,
    with DeltaHV with gap, with DeltaHV total,
    with DeltaHViPi with gap, with DeltaHViPi total}
  */
  
  /*   Converting deductor event to FastJet format
   *   as defined in fastjet.h
   */
  make_pseudojets(theparticles, p);
  
  /*    Run fastjet algorithm   */
  fj::ClusterSequence clustSeq(theparticles, jetDef);
  
  /*    Extract inclusive jets sorted by pT (above pTcut)   */
  double pTcut = 20.0_GeV;
  incJets0 = clustSeq.inclusive_jets(pTcut);
  /* Sort the jets */
  incJets = sorted_by_pt(incJets0);
  
  // Find two leading PT jets in allowed rapidity range.
  fj::PseudoJet jet[2];
  unsigned int nj = 0;
  for(unsigned int n = 0; n < incJets.size() && nj < 2; n++)
    if(std::abs(incJets[n].rapidity()) < 4.4) jet[nj++] = incJets[n];
  if(nj < 2) return;
  
  double Ymin = jet[0].rapidity(), Ymax = jet[1].rapidity();
  double DeltaY = std::abs(Ymax - Ymin);
  if(Ymin > Ymax) std::swap(Ymin, Ymax);
  double pTaverage = 0.5*(jet[0].perp() + jet[1].perp());
  
  for(unsigned int i = 0; i < incJets.size(); i++)
    if(incJets[i].pt() > pTcut &&
       incJets[i].rapidity() < Ymax - tiny &&
       incJets[i].rapidity() > Ymin + tiny )
    {
      w[0] = 0.0; // There is no gap.
      w[2] = 0.0; // There is no gap.
      w[4] = 0.0; // There is no gap.
      break;
    }
  
  if(1.0 < DeltaY && DeltaY < 2.0)
  {
    double cs = cross_section12(pTaverage);
    for(unsigned int i = 0; i < 6; i++) w[i] = w[i]/cs;
    HistArr::physfill(1, w, dirac(), pTaverage);
  }
  
  if (2.0 < DeltaY && DeltaY < 3.0)
  {
    double cs = cross_section23(pTaverage);
    for(unsigned int i = 0; i < 6; i++) w[i] = w[i]/cs;
    HistArr::physfill(2, w, dirac(), pTaverage);
  }
  
  if(3.0 < DeltaY && DeltaY < 4.0)
  {
    double cs = cross_section34(pTaverage);
    for(unsigned int i = 0; i < 6; i++) w[i] = w[i]/cs;
    HistArr::physfill(3, w, dirac(), pTaverage);
  }
  
  if (4.0 < DeltaY && DeltaY < 5.0)
  {
    double cs = cross_section45(pTaverage);
    for(unsigned int i = 0; i < 6; i++) w[i] = w[i]/cs;
    HistArr::physfill(4, w, dirac(), pTaverage);
  }
  
  if(5.0 < DeltaY && DeltaY < 6.0)
  {
    double cs = cross_section56(pTaverage);
    for(unsigned int i = 0; i < 6; i++) w[i] = w[i]/cs;
    HistArr::physfill(5, w, dirac(), pTaverage);
  }
  
} // End void UserJets::userfunc(...)

